# coding: utf-8
u"""Модели для хранения данных системы авторизации RBAC."""
from django.contrib.contenttypes.generic import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.fields import FieldDoesNotExist
from django.db.models.signals import post_delete
from django.dispatch import receiver

from educommon.django.db.mixins.date_interval import DateIntervalMixin


class _BaseModel(models.Model):

    u"""Базовый класс для моделей приложения."""

    def __unicode__(self):
        return u'{}{{{}}}'.format(self.__class__.__name__, self.pk)

    def full_clean(self, exclude=None):
        self._full_clean_was_called = True

        super(_BaseModel, self).full_clean(exclude=exclude)

    def save(self, *args, **kwargs):
        assert getattr(self, '_full_clean_was_called', False), (
            u'Attempt to save data without validation (model {}.{})'.format(
                self.__class__.__module__, self.__class__.__name__
            )
        )

        for field in self._meta.fields:
            if (isinstance(field, (models.TextField, models.CharField)) and
                    not getattr(self, field.attname)):
                setattr(self, field.attname, None if field.null else u'')

        super(_BaseModel, self).save(*args, **kwargs)

    class Meta:
        abstract = True


class Permission(_BaseModel):

    u"""Разрешение."""

    name = models.CharField(
        u'Имя',
        max_length=100,
        db_index=True,
        unique=True,
    )
    title = models.CharField(
        u'Название',
        max_length=200,
        unique=True,
        blank=True, null=True,
    )
    description = models.TextField(
        u'Описание',
        blank=True,
    )

    class Meta:
        verbose_name = u'Разрешение'
        verbose_name_plural = u'Разрешения'

    def __unicode__(self):
        return u'Permission<{}: {}>'.format(self.id, self.name)


class Role(_BaseModel):

    u"""Роль."""

    name = models.CharField(
        u'Имя',
        max_length=300,
        db_index=True,
        unique=True,
    )
    description = models.TextField(
        u'Описание',
        blank=True,
    )
    permissions = models.ManyToManyField(
        Permission,
        related_name='roles',
    )

    class Meta:
        verbose_name = u'Роль'
        verbose_name_plural = u'Роли'

    def __unicode__(self):
        return u'Role<{}: {}>'.format(self.id, self.name)

    @property
    def subroles(self):
        u"""Возвращает все вложенные роли данной роли.

        :rtype: set
        """
        result = set()

        for rp in RoleParent.objects.filter(parent_id=self.id):
            result.add(rp.role)
            result.update(rp.role.subroles)

        return result

    def get_permissions(self):
        u"""Возвращает все разрешения роли, в т.ч. вложенных ролей.

        :rtype: QuerySet
        """
        m2m_model = self._meta.get_field('permissions').rel.through

        roles = set([self]) | self.subroles
        result = Permission.objects.filter(
            pk__in=m2m_model.objects.filter(
                role__in=roles,
            ).distinct('permission').values('permission'),
        )

        return result


class RoleParent(models.Model):

    u"""M2M-модель "Вложенная роль"."""

    parent = models.ForeignKey(Role, related_name='+')
    role = models.ForeignKey(Role, related_name='+')

    def clean(self):
        errors = {}

        try:
            super(RoleParent, self).clean()
        except ValidationError as error:
            errors = error.update_error_dict(errors)

        if self.parent.id == self.role.id:
            errors.setdefault('parent', []).append(
                u'Роль не может содержать сама себя'
            )

        # ---------------------------------------------------------------------
        # Проверка отсутствия цикла
        query = RoleParent.objects.all()
        if self.pk:
            query = query.exclude(pk=self.pk)

        def check(target_role, role):
            for rp in query.filter(role=role):
                if target_role.id == rp.parent_id:
                    raise ValidationError(u'В иерархии ролей обнаружен цикл')
                check(target_role, rp.parent)

        try:
            # Проверка, нет ли self.role среди предков self.parent
            check(self.role, self.parent)
        except ValidationError as error:
            errors.setdefault('parent', []).extend(error.messages)
        # ---------------------------------------------------------------------

        if errors:
            raise ValidationError(errors)

    def __unicode__(self):
        return u'RoleParent({} --> {})'.format(
            unicode(self.role), unicode(self.parent)
        )

    class Meta:
        unique_together = ('parent', 'role')


class UserRole(DateIntervalMixin, models.Model):

    u"""M2M-модель "Роль пользователя"."""

    no_intersections_for = ('user', 'role')

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField()
    user = GenericForeignKey()
    role = models.ForeignKey(Role)

    def __unicode__(self):
        return u'UserRole({} --> {})'.format(
            unicode(self.user), unicode(self.role),
        )

    class Meta:
        unique_together = ('content_type', 'object_id', 'role')


@receiver(post_delete)
def delete_user_roles(instance, **kwargs):
    u"""Удаление привязки ролей к пользователям при удалении пользователя."""
    content_type = ContentType.objects.get_for_model(instance)

    try:
        content_type.model_class()._meta.get_field_by_name('id')
    except FieldDoesNotExist:
        return

    UserRole.objects.filter(
        content_type=content_type,
        object_id=instance.id,
    ).delete()
